/////////////////////////////////////////////////////////////////////////////
//           WWWW       WWW                                      
//            WW         W                                       
//            WW         W                                       
//            WW         W                                       
//            WW    W    W     WWWWWW    WWWW    WWW     WWWWW   
//            WW   WWW   W    WW    WW    WW      W    WW    WWW 
//            WW  W  WW  W          WW    WW      W   WW     WW  
//            WW W    WW W    WWWWWWWW     WW    W    WW   WW    
//            WWW      WWW   WW     WW      WW  W     WWWWW      
//            WW        WW   WW    WWW       WWW       WW     WW 
//            W          W    WWWWW WWW       W          WWWWW   
//                                                    
//                                                    
//                                                    
//       WW        WW                                      
//       WWW      WWW                  W                   
//       W WW    W WW                 WW     WW            
//       W  WW  W  WW                 WW                   
//       W   WWW   WW      WWWW     WWWWWW   WW      WWWW     WWW WWWWW   
//       W    W    WW    WW    WW     WW    WWW    WW    WW    WWW    WW  
//       W         WW   WW      WW    WW     WW   WW      WW   WW     WW  
//       W         WW   WW      WW    WW     WW   WW      WW   WW     WW  
//       W         WW   WW      WW    WW     WW   WW      WW   WW     WW  
//       W         WW    WW    WW     WW     WW    WW    WW    WW     WW  
//      WWW       WWWW     WWWW        WWW  WWWW     WWWW     WWWW   WWWW 
//                                                                          
//                                                                          
//                                                                         
//   WWWWW                                 WWW                               
// WW     WW                                WW              W                 
//WW       W   WW                           WW             WW                 
//WW                                        WW             WW                 
// WWW         WW  WWW WW  WW   WWW   WWW   WW    WWWWW  WWWWWW   WWW    WWW WWW
//   WWWWWW   WWW   WWW WWW WW   WW    WW   WW   WW   WW   WW   WW   WW   WWW   
//       WWW   WW   WW  WW  WW   WW    WW   WW        WW   WW  WW     WW  WW    
//         WW  WW   WW  WW  WW   WW    WW   WW   WWWWWWW   WW  WW     WW  WW    
//W        WW  WW   WW  WW  WW   WW    WW   WW  WW    WW   WW  WW     WW  WW    
//WW      WW   WW   WW  WW  WW   WW   WWW   WW  WW   WWW   WW   WW   WW   WW    
//  WWWWWW    WWWW WWWW WW WWWW   WWWW WWW WWWW  WWWW WWW   WWW   WWW    WWWW   
//                                                                        
//                                                                        
 
 
 

////////////////////////////////////////////////////////////
//
// waveSimulator.cxx 
// Display a surface mesh with waves produced by an oscillator.
// Version 1.00
//
// Copyright (c) 2002   David C. Banks  
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public Licence
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
// banks@cs.fsu.edu
//
// Dr. David C. Banks
// Department of Computer Science
// Florida State University
// 488 Dirac Science Library 
// Tallahassee, FL 32306
//

/////////////////////////////////////////////////////////////////////////////
//                                                                            
// WWWWW                                                                      
// WW  WW                              WW          WW   WW                    
// WW  WW                                          WW                         
// WW  WW   WWW    WWWW   WWW   WW WW  WW  WWWW   WWWW  WW   WWW   WWWW       
// WW  WW  WW WW  WW     WW WW  WWW    WW  WW WW   WW   WW  WW WW  WW WW      
// WW  WW  WWWWW   WWW   WW     WW     WW  WW WW   WW   WW  WW WW  WW WW      
// WW  WW  WW        WW  WW WW  WW     WW  WW WW   WW   WW  WW WW  WW WW      
// WWWWW    WWWW  WWWW    WWW   WW     WW  WWWW    WW   WW   WWW   WW WW      
//                                         WW                                 
//                                         WW                                 


////////////////////////////////////////////////////////////
//
// Description
//
// This program simulates waves moving on a surface.
//
// The height z is computed by evaluating a function of x, y,
// and time. The function models the waves generated by a periodic
// oscillator in a ripple tank. The oscillator has parameters
//
//   position
//   amplitude
//   frequency
//   phase
//
// The ripple tank has parameters
//
//   waveSpeed
//   damping
//


/////////////////////////////////////////////////////////////////////////////
//                                                                           
// WWWW                WWW             WW                                    
//  WW                  WW             WW                                    
//  WW                  WW             WW                                    
//  WW   WWWW    WWW    WW   WW WW   WWWW   WWW    WWWW                      
//  WW   WW WW  WW WW   WW   WW WW  WW WW  WW WW  WW                         
//  WW   WW WW  WW      WW   WW WW  WW WW  WWWWW   WWW                       
//  WW   WW WW  WW WW   WW   WW WW  WW WW  WW        WW                      
// WWWW  WW WW   WWW   WWWW   WWWW   WWWW   WWWW  WWWW                       
//                                                                           
// Includes

#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#include <Inventor/Xt/SoXt.h>
#include <Inventor/Xt/viewers/SoXtExaminerViewer.h>

#include <Inventor/nodes/SoCamera.h>
#include <Inventor/nodes/SoSeparator.h>

#include <Inventor/sensors/SoTimerSensor.h>

#include "grid.h"
#include "oscillator.h"
#include "rippleTank.h"
#include "lookup.h"


/////////////////////////////////////////////////////////////////////////////
//                                                                            
//  WWWWW  WWW          WW             WWW                                    
// WW  WW   WW          WW              WW                                    
// WW       WW          WW              WW                                    
// WW WWW   WW    WWW   WWWW    WWWW    WW    WWWW                            
// WW  WW   WW   WW WW  WW WW      WW   WW   WW                               
// WW  WW   WW   WW WW  WW WW   WWWWW   WW    WWW                             
// WW  WW   WW   WW WW  WW WW  WW  WW   WW      WW                            
//  WWWWW  WWWW   WWW   WWWW    WWWWW  WWWW  WWWW                             
//                                                                            
// Globals



// Global Variables

        GridStruct       Grid;
	RippleTankStruct RippleTank;
	OscillatorStruct Oscillator;

        float            SimulationTime;
        double           TimeIncrement;

        SoTimerSensor   *TimerSensor;
        LookupTable     L;
        int             OptShowTimes = 0;
        int             OptUseLookups = 1;


// Function declarations



SoSeparator *
createScene();

void
updateMesh(int gridpointBudget);

SoXtExaminerViewer *
createAndLayoutMainWindow(int argc, char **argv);




/////////////////////////////////////////////////////////////////////////////
//                                                                            
// WWWW                               WWW                                     
//  WW          WW   WW   WW           WW   WW                                
//  WW               WW                WW                                     
//  WW   WWWW   WW  WWWW  WW   WWWW    WW   WW  WWWWW   WWW                   
//  WW   WW WW  WW   WW   WW      WW   WW   WW    WW   WW WW                  
//  WW   WW WW  WW   WW   WW   WWWWW   WW   WW   WW    WWWWW                  
//  WW   WW WW  WW   WW   WW  WW  WW   WW   WW  WW     WW                     
// WWWW  WW WW  WW   WW   WW   WWWWW  WWWW  WW  WWWWW   WWWW                  
//                                                                            
//                                                                            

////////////////////////////////////////////////////////////
//
// Initialize the state of the simulation.
//

void
initializeSimulation ()
  {
  Oscillator.position           = SbVec2f(0.0, 0.0);
  Oscillator.velocity           = SbVec2f(0.6, 0.0);
  Oscillator.amplitude          = 0.1;
  Oscillator.frequency          = 0.5;
  Oscillator.phase              = 0.0;


  RippleTank.waveSpeed          = 1.0;
  RippleTank.damping            = 1.0;



  // The grid size and time increments are chosen to
  // give real-time performance on a system that can
  // render the mesh while updating the positions of its vertices. 
  // Assuming dim = sDim = tDim (a square mesh), there are
  //  
  //    (dim-1)^2 * 2 triangles
  //
  // in the quadrilateral mesh, and
  //
  //    dim^2 grid points.
  //
  // Updating a grid point is dominated by the sin(), exp(), and
  // sqrt() evaluations, which may each require some 250 machine 
  // instructions.  If about 1000 CPU instructions are used 
  // (sin + exp + sqrt + miscellaneous) in evaluating a grid point, 
  // the CPU requirement  m   is
  // 
  //     m Hz   >   1000 * dim^2 / timeIncrement 
  //
  // and the graphics requirement is
  //
  //     s triangle/sec   >  (dim-1)^2 * 2 / timeIncrement
  //
  // For a mesh with  dim = 200 and displaying at timeIncrement = 0.05 
  // (i.e. 20 frames/sec), we need
  //
  //     m      >   800,000,000 floatingPointOperation/sec
  //
  // which is 800 megaFLOPs.  A 2,000,000,000 Hz (2 gigaHertz) processor 
  // is typical for a desktop computer in year 2002, permitting about 
  // 3 clock cycles for each floating-point operation.
  // 
  // For the graphics subsystem we need
  //
  //     s      >   1,584,040 triangle/sec
  // 
  // A 2,000,000 triangle/sec graphics card is typical for
  // a desktop computer in year 2002 (nVidia GeForce 4).

  Grid.sDim                   =  200;
  Grid.tDim                   =  200;

  Grid.xBounds                =  SbVec2f(-1.0, 1.0);
  Grid.yBounds                =  SbVec2f(-1.0, 1.0);

  Grid.resolutionChanged      =  TRUE;


  SimulationTime              =    0.0;
  TimeIncrement               =    0.10;



  //Grid.gridpointBudget        =  Grid.sDim*Grid.tDim*0.11;
  Grid.gridpointBudget          =  Grid.sDim*Grid.tDim;
  }








/////////////////////////////////////////////////////////////////////////////
//                                                                           
//   WW                                                                      
//  WWWW          WW                     WW                                  
// WW  WW                                WW                                  
// WW  WW  WWWW   WW  WWWW WW    WWWW   WWWW   WWW                           
// WWWWWW  WW WW  WW  WW WW WW      WW   WW   WW WW                          
// WW  WW  WW WW  WW  WW WW WW   WWWWW   WW   WWWWW                          
// WW  WW  WW WW  WW  WW WW WW  WW  WW   WW   WW                             
// WW  WW  WW WW  WW  WW WW WW   WWWWW   WW    WWWW                          
//                                                                           
//                                                                           
// Update the simulation.
// This is a continual scheduled update, called by the
// Inventor main loop.

void
animate(void *viewer, SoSensor *)
  {
  updateMesh(Grid.gridpointBudget);

  }



/////////////////////////////////////////////////////////////////////////////
//                                                                            
//                                       WW                                   
//                                      WWWW                                  
//                                     WW  WW                                 
// WWWW    WWWW   WW WW   WWWW   WWW   WW  WW  WW WW   WWWW  WW WW  WWWW WW   
// WW WW      WW  WWW    WW     WW WW  WWWWWW  WWW    WW WW  WW WW  WW WW WW  
// WW WW   WWWWW  WW      WWW   WWWWW  WW  WW  WW     WW WW  WW WW  WW WW WW  
// WW WW  WW  WW  WW        WW  WW     WW  WW  WW     WW WW  WW WW  WW WW WW  
// WWWW    WWWWW  WW     WWWW    WWWW  WW  WW  WW      WWWW   WWWW  WW WW WW  
// WW                                                    WW                   
// WW                                                 WWWW                    

void parseArguments(int argc, char **argv){
  int i;
  for (i=1; i<argc; i++){
    if (strcmp(argv[i], "-h") == 0 ||
	strcmp(argv[i], "--help") == 0){
      printf("Usage: %s [OPTIONS]\n\n", argv[0]);
      printf("\t--help\t\tShow this help message.\n");
      printf("\t--showtimes\tShow number of CPU cycles for each frame.\n");
      printf("\t--nolookups\tDon't use function lookups (slower).\n\n\n");
      exit(0);
    }
    else if (strcmp(argv[i], "--showtimes") == 0){
      OptShowTimes++;
    }
    else if (strcmp(argv[i], "--nolookups") == 0){
      OptUseLookups=0;
    }
  }
}


/////////////////////////////////////////////////////////////////////////////
//                                                  WW    WW                  
//                                                 W        W                 
//                              WW                WW        WW                
//                                                WW        WW                
// WWW WWW  WWW      WWWWWW     WW   WWW WWWWW    WW        WW                
//  WWW  WWW  WW    WW    WW   WWW    WWW    WW   WW        WW                
//  WW   WW   WW          WW    WW    WW     WW   WW        WW                
//  WW   WW   WW    WWWWWWWW    WW    WW     WW   WW        WW                
//  WW   WW   WW   WW     WW    WW    WW     WW   WW        WW                
//  WW   WW   WW   WW    WWW    WW    WW     WW   WW        WW                
// WWWW WWWW WWWW   WWWWW WWW  WWWW  WWWW   WWWW   W        W                 
//                                                  WW    WW                  
//                                                                            
//                                                                            


////////////////////////////////////////////////////////////
//
// main
//
// Initialize global variables.
// Create the graphics window.
// Create the scene graph.
// Schedule a periodic update to the scene graph.
// Give control to the Inventor main loop.

void
main(int argc, char **argv)
  {
  parseArguments(argc, argv);
    
  initializeSimulation();
  L.calcTable();

  SoXtExaminerViewer *viewer = 
    createAndLayoutMainWindow(argc, argv);


  SoSeparator *root = createScene();
  updateMesh(Grid.sDim*Grid.tDim);

  TimerSensor = new SoTimerSensor(animate, viewer);
  TimerSensor->setInterval(SbTime(TimeIncrement)); // seconds per frame
  TimerSensor->schedule();


  viewer->setSceneGraph(root);


  // INVENTOR HACK:
  //
  // We have put a cube-environment around the scene, which causes
  // the Viewer's camera to be positioned far enough back to
  // view the entire cube. In order to bring the camera close
  // the the mesh inside the cube, we intervene to place
  // its position closer to the center of the mesh (0.0, 0.0, 0.0) .
  // 
  // Having done that, we re-set the camera's focal distance
  // so that interactive rotation with the mouse will rotate about
  // the center of the mesh, rather than rotating about a center
  // still far away. Then we save this setting as the home-position of
  // the Viewer's camera.

  double cameraDistance = 2.0;

  SoCamera *camera = viewer->getCamera();
  SbVec3f position = camera->position.getValue();
  position.normalize();
  position *= cameraDistance;
  camera->position.setValue(position);
  camera->focalDistance = cameraDistance;
  viewer->saveHomePosition();
  viewer->setBackgroundColor(SbColor(1.0, 0.0, 0.0));

  viewer->show();

  // Hand over control to Inventor.
  SoXt::mainLoop();
  }
